from dataclasses import dataclass
from typing import Any

import sqlalchemy as sa
from sqlalchemy.engine.base import Engine as SaEngine
from sqlalchemy.engine.row import RowProxy

from spinta.components import Property
from spinta.core.ufuncs import Env
from spinta.utils.data import take
from spinta.utils.schema import NA


@dataclass
class Engine:
    dsn: str  # sqlalchemy engine dsn
    schema: str = None
    encoding: str = NA

    def create(self) -> SaEngine:
        return sa.create_engine(self.dsn, **take({
            'encoding': self.encoding,
        }))


class SqlResource(Env):
    dsn: str

    def init(self, dsn: str):
        return self(dsn=dsn)


class SqlResultBuilder(Env):
    this: Any           # A selected value.
    prop: Property      # Model property.
    data: RowProxy      # Whole row from database.

    def init(self, this: Any, prop: Property, data: RowProxy):
        return self(
            this=this,
            prop=prop,
            data=data,
        )



import os
import shutil
import uuid
import ruamel.yaml
from typing import Type

import pkg_resources as pres
import logging

from authlib.common.errors import AuthlibHTTPError
from authlib.oauth2.rfc6750.errors import InsufficientScopeError

from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.responses import RedirectResponse
from starlette.templating import Jinja2Templates
from starlette.staticfiles import StaticFiles
from starlette.routing import Route, Mount
from starlette.middleware import Middleware

from spinta import components, commands
from spinta.auth import AuthorizationServer, check_scope, query_client, get_clients_list, \
    client_exists, create_client_file, delete_client_file, update_client_file, get_clients_path
from spinta.auth import ResourceProtector
from spinta.auth import BearerTokenValidator
from spinta.auth import get_auth_request
from spinta.auth import get_auth_token
from spinta.commands import prepare, get_version
from spinta.components import Context
from spinta.exceptions import BaseError, MultipleErrors, error_response, InsufficientPermission, \
    UnknownPropertyInRequest, InsufficientPermissionForUpdate, EmptyPassword
from spinta.middlewares import ContextMiddleware
from spinta.urlparams import Version
from spinta.urlparams import get_response_type
from spinta.utils.response import create_http_response
from spinta.accesslog import create_accesslog
from spinta.exceptions import NoAuthServer

log = logging.getLogger(__name__)

templates = Jinja2Templates(
    directory=pres.resource_filename('spinta', 'templates'),
)


async def favicon(request: Request):
    return Response('', media_type='text/plain', status_code=404)


async def robots(request: Request):
    content = '\n'.join([
        'User-Agent: *',
        'Allow: /',
        '',
    ])
    return Response(content, media_type='text/plain', status_code=200)


async def version(request: Request):
    version = get_version(request.state.context)
    return JSONResponse(version)


async def auth_token(request: Request):
    auth_server = request.state.context.get('auth.server')
    if auth_server.enabled():
        return auth_server.create_token_response({
            'method': request.method,
            'url': str(request.url.replace(query='')),
            'body': dict(await request.form()),
            'headers': request.headers,
        })
    else:
        raise NoAuthServer()


def _auth_client_context(request: Request) -> Context:
    context: Context = request.state.context
    context.set('auth.request', get_auth_request({
        'method': request.method,
        'url': str(request.url.replace(query='')),
        'body': None,
        'headers': request.headers,
    }))
    context.set('auth.token', get_auth_token(context))
    context.attach('accesslog', create_accesslog, context, loaders=(
        context.get('store'),
        context.get("auth.token"),
    ))
    return context


async def auth_clients_add(request: Request):
    try:
        context = _auth_client_context(request)
        check_scope(context, 'auth_clients')
        config = context.get('config')
        commands.load(context, config)

        path = get_clients_path(config)
        data = await request.json()
        for key in data.keys():
            if key not in ('client_name', 'secret', 'scopes'):
                raise UnknownPropertyInRequest(property=key, properties=('client_name', 'secret', 'scopes'))
        client_id = str(uuid.uuid4())
        name = data["client_name"] if ("client_name" in data.keys() and data["client_name"]) else client_id

        while client_exists(path, client_id):
            client_id = str(uuid.uuid4())

        if "secret" in data.keys() and not data["secret"] or "secret" not in data.keys():
            raise EmptyPassword

        client_file, client_ = create_client_file(
            path,
            name,
            client_id,
            data["secret"],
            data["scopes"] if "scopes" in data.keys() else None,
        )

        return JSONResponse({
            "client_id": client_["client_id"],
            "client_name": name,
            "scopes": client_["scopes"]
        })

    except InsufficientScopeError:
        raise InsufficientPermission(scope='auth_clients')


async def auth_clients_get_all(request: Request):
    try:
        context = _auth_client_context(request)
        check_scope(context, 'auth_clients')
        config = context.get('config')
        commands.load(context, config)

        path = get_clients_path(config)
        ids = get_clients_list(path)
        return_values = []
        for client_path in ids:
            client = query_client(path, client_path)
            return_values.append({
                "client_id": client.id,
                "client_name": client.name
            })
        return JSONResponse(return_values)

    except InsufficientScopeError:
        raise InsufficientPermission(scope='auth_clients')


async def auth_clients_get_specific(request: Request):
    try:
        context = _auth_client_context(request)
        token = context.get('auth.token')
        config = context.get('config')
        commands.load(context, config)

        client_id = request.path_params["client"]
        if client_id != token.get_client_id():
            check_scope(context, 'auth_clients')

        path = get_clients_path(config)
        client = query_client(path, client_id)
        return_value = {
            "client_id": client.id,
            "client_name": client.name,
            "scopes": list(client.scopes)
        }
        return JSONResponse(return_value)

    except InsufficientScopeError:
        raise InsufficientPermission(scope='auth_clients')


async def auth_clients_delete_specific(request: Request):
    try:
        context = _auth_client_context(request)
        client_id = request.path_params["client"]
        check_scope(context, 'auth_clients')
        config = context.get('config')
        commands.load(context, config)

        path = get_clients_path(config)
        delete_client_file(path, client_id)
        return Response(status_code=204)

    except InsufficientScopeError:
        raise InsufficientPermission(scope='auth_clients')


async def auth_clients_patch_specific(request: Request):
    try:
        context = _auth_client_context(request)
        token = context.get('auth.token')
        client_id = request.path_params["client"]

        config = context.get('config')
        commands.load(context, config)
        path = get_clients_path(config)

        has_permission = True
        try:
            check_scope(context, 'auth_clients')
        except InsufficientScopeError:
            has_permission = False
        if client_id != token.get_client_id() and not has_permission:
            raise InsufficientScopeError

        data = await request.json()
        for key in data.keys():
            if key not in ('client_name', 'secret', 'scopes'):
                properties = ('secret')
                if has_permission:
                    properties = ('client_name', 'secret', 'scopes')
                raise UnknownPropertyInRequest(property=key, properties=properties)
            elif key != 'secret' and not has_permission:
                raise InsufficientPermissionForUpdate(field=key)

        if "secret" in data.keys() and not data["secret"]:
            raise EmptyPassword

        new_data = update_client_file(
            context,
            path,
            client_id,
            name=data["client_name"] if "client_name" in data.keys() and data["client_name"] else None,
            secret=data["secret"] if "secret" in data.keys() and data["secret"] else None,
            scopes=data["scopes"] if "scopes" in data.keys() and data["scopes"] is not None else None,
        )
        return_value = {
            "client_id": new_data["client_id"],
            "client_name": new_data["client_name"],
            "scopes": list(new_data["scopes"])
        }
        return JSONResponse(return_value)

    except InsufficientScopeError:
        raise InsufficientPermission(scope='auth_clients')


async def homepage(request: Request):
    context: Context = request.state.context

    context.set('auth.request', get_auth_request({
        'method': request.method,
        'url': str(request.url.replace(query='')),
        'body': None,
        'headers': request.headers,
    }))
    context.set('auth.token', get_auth_token(context))

    config = context.get('config')
    UrlParams: Type[components.UrlParams]
    UrlParams = config.components['urlparams']['component']
    params: UrlParams = prepare(context, UrlParams(), Version(), request)

    context.attach('accesslog', create_accesslog, context, loaders=(
        context.get('store'),
        context.get("auth.token"),
        request,
        params,
    ))

    return await create_http_response(context, params, request)


async def error(request, exc):
    log.exception('Error: %s', exc)

    headers = {}

    if isinstance(exc, MultipleErrors):
        # TODO: probably there should be a more sophisticated way to get status
        #       code from error list.
        # XXX:  On the other hand, mixing status code is probably not a good
        #       idea, so instead, MultipleErrors must have one status code for
        #       all errors. And that should be enforced by MultipleErrors. If
        #       an exception comes in with a different status code, then this
        #       should be an error.
        status_code = exc.errors[0].status_code
        errors = [error_response(e) for e in exc.errors]

    elif isinstance(exc, BaseError):
        status_code = exc.status_code
        errors = [error_response(exc)]
        headers = exc.headers
    else:
        if isinstance(exc, HTTPException):
            status_code = exc.status_code
            message = exc.detail
        elif isinstance(exc, AuthlibHTTPError):
            status_code = exc.status_code
            if exc.description:
                # show overridden description
                message = str(f'{exc.error}: {exc.description}')
            elif exc.get_error_description():
                # show dynamic description
                message = str(f'{exc.error}: {exc.get_error_description()}')
            else:
                # if no description, show plain error string
                message = exc.error
            headers['www-authenticate'] = f'Bearer error="{exc.error}"'
        else:
            status_code = 500
            message = str(exc)

        errors = [
            {
                'code': type(exc).__name__,
                'message': message,
            }
        ]

    response = {'errors': errors}

    fmt = get_response_type(request.state.context, request)
    if fmt == 'json':
        return JSONResponse(
            response,
            status_code=status_code,
            headers=headers,
        )
    else:
        response = {
            **response,
            'request': request,
        }
        return templates.TemplateResponse(
            'error.html',
            response,
            status_code=status_code,
            headers=headers,
        )


async def srid_check(request: Request):
    from shapely.geometry import Point
    from geoalchemy2.shape import from_shape
    from spinta.backends.postgresql.types.geometry.helpers import get_osm_link

    srid = request.path_params['srid']
    x = request.path_params['x']
    y = request.path_params['y']

    point = Point(x, y)
    wkb = from_shape(point, srid)
    osm_link = get_osm_link(wkb, srid)
    return RedirectResponse(url=osm_link)


def init(context: Context):
    config = context.get('config')

    routes = [
        Route('/robots.txt', robots, methods=['GET']),
        Route('/favicon.ico', favicon, methods=['GET']),
        Route('/version', version, methods=['GET']),
        Route('/auth/token', auth_token, methods=['POST']),
        Route('/_srid/{srid:int}/{x:float}/{y:float}', srid_check, methods=['GET']),
        Route('/auth/clients', auth_clients_get_all, methods=['GET']),
        Route('/auth/clients', auth_clients_add, methods=['POST']),
        Route('/auth/clients/{client}', auth_clients_get_specific, methods=['GET']),
        Route('/auth/clients/{client}', auth_clients_delete_specific, methods=['DELETE']),
        Route('/auth/clients/{client}', auth_clients_patch_specific, methods=['PATCH'])
    ]

    if config.docs_path:
        routes += [
            Mount('/docs', StaticFiles(directory=config.docs_path, html=True)),
        ]

    # This route matches everything, so it must be added last.
    routes += [
        Route('/{path:path}', homepage, methods=[
            'HEAD',
            'GET',
            'POST',
            'PUT',
            'PATCH',
            'DELETE'
        ]),
    ]

    middleware = [
        Middleware(ContextMiddleware, context=context)
    ]

    exception_handlers = {
        Exception: error,
        BaseError: error,
        MultipleErrors: error,
        AuthlibHTTPError: error,
        HTTPException: error,
    }

    app = Starlette(
        debug=config.debug,
        routes=routes,
        middleware=middleware,
        exception_handlers=exception_handlers,
    )

    # Add context to state in order to pass it to request handlers
    app.state.context = context

    context.bind('auth.server', AuthorizationServer, context)
    context.bind(
        'auth.resource_protector',
        ResourceProtector,
        context,
        BearerTokenValidator,
    )

    return app



import logging
import pathlib
import types
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import tqdm

from spinta import commands
from spinta import components
from spinta.cli.helpers.errors import ErrorCounter
from spinta.formats.components import Format
from spinta.components import Context
from spinta.components import DataStream
from spinta.components import Model
from spinta.core.ufuncs import Expr
from spinta.ufuncs.basequerybuilder.components import QueryParams
from spinta.ufuncs.basequerybuilder.ufuncs import add_page_expr
from spinta.types.datatype import Inherit
from spinta.utils.aiotools import alist
from spinta.utils.itertools import peek

ModelRow = Tuple[Model, Dict[str, Any]]

log = logging.getLogger(__name__)


def _get_row_count(
    context: components.Context,
    model: components.Model,
) -> int:
    query = Expr('select', Expr('count'))
    if model.page.is_enabled:
        copied = deepcopy(model.page)
        copied.filter_only = True
        query = add_page_expr(query, copied)
    stream = commands.getall(context, model, model.backend, query=query)
    for data in stream:
        return data['count()']


def count_rows(
    context: Context,
    models: List[Model],
    limit: int = None,
    *,
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None
) -> Dict[str, int]:
    counts = {}
    for model in tqdm.tqdm(models, 'Count rows', ascii=True, leave=False):
        try:
            count = _get_row_count(context, model)
        except Exception:
            if error_counter:
                error_counter.increase()
                if error_counter.has_reached_max():
                    break
            if stop_on_error:
                break
            else:
                log.exception("Error on _get_row_count({model.name}).")
        else:
            if limit:
                count = min(count, limit)
            counts[model.name] = count
    return counts


def read_model_data(
    context: components.Context,
    model: components.Model,
    limit: int = None,
    stop_on_error: bool = False,
    params: QueryParams = None
) -> Iterable[Dict[str, Any]]:

    if limit is None:
        query = None
    else:
        query = Expr('limit', limit)

    stream = commands.getall(context, model, model.backend, query=query, params=params)

    if stop_on_error:
        stream = peek(stream)
    else:
        try:
            stream = peek(stream)
        except Exception:
            log.exception(f"Error when reading data from model {model.name}")
            return

    prop_filer, needs_filtering = filter_allowed_props_for_model(model)
    for item in stream:
        if needs_filtering:
            item = filter_dict_by_keys(prop_filer, item)
        yield item


def filter_allowed_props_for_model(model: Model) -> (dict, bool):
    if model.base:
        allowed_props = model.properties
        for name, prop in model.base.parent.properties.items():
            if not name.startswith('_'):
                if name in allowed_props and isinstance(allowed_props[name].dtype, Inherit):
                    allowed_props.pop(name)
        return allowed_props, True
    return model.properties, False


def filter_dict_by_keys(dict1: dict, dict2: dict) -> dict:
    result = {}
    for key in dict1:
        if key in dict2:
            result[key] = dict2[key]
    return result


def iter_model_rows(
    context: Context,
    models: List[Model],
    counts: Dict[str, int],
    limit: int = None,
    *,
    stop_on_error: bool = False,
    no_progress_bar: bool = False,
) -> Iterator[ModelRow]:
    for model in models:
        rows = read_model_data(context, model, limit, stop_on_error)
        if not no_progress_bar:
            count = counts.get(model.name)
            rows = tqdm.tqdm(rows, model.name, ascii=True, total=count, leave=False)
        for row in rows:
            yield model, row


async def process_stream(
    source: str,
    stream: DataStream,
    exporter: Optional[Format] = None,
    path: Optional[pathlib.Path] = None,
) -> None:
    if exporter:
        # TODO: Probably exporters should support async generators.
        if isinstance(stream, types.AsyncGeneratorType):
            stream = await alist(stream)
        chunks = exporter(stream)
        if path is None:
            for chunk in chunks:
                print(chunk, end='')
        else:
            with path.open('wb') as f:
                for chunk in chunks:
                    f.write(chunk)
    else:
        with tqdm.tqdm(desc=source, ascii=True) as pbar:
            async for _ in stream:
                pbar.update(1)


def ensure_data_dir(
    path: pathlib.Path,
):
    """Create default data directory if it does not exits"""
    path.mkdir(parents=True, exist_ok=True)



from __future__ import annotations

from typing import List
from typing import Optional

from spinta.components import Property
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import Unresolved
from spinta.exceptions import IncompatibleForeignProperties
from spinta.types.datatype import DataType
from spinta.types.datatype import Ref


class ForeignProperty(Unresolved):
    """Representation of a reference.

    When querying `/City?select(country.code)`, `left` points to `country` and
    `right` points to `code`. `chain` will be:

        [
            ForeignProperty(City, country:Country -> code:String),
        ]

    Multiple references can be joined like this:

        /City?select(country.continent.planet.name)

    `chain` will look like this:

        [
            ForeignProperty(
                City,
                country:Country ->
                continent:ref
            ),
            ForeignProperty(
                City,
                country:Country ->
                continent:Continent ->
                planet:ref
            ),
            ForeignProperty(
                City,
                country:Country ->
                continent:Continent ->
                planet:Planet ->
                name:string
            ),
        ]

    """

    left: Ref
    right: Optional[DataType]
    chain: List[ForeignProperty]
    name: str

    def __init__(
        self,
        fpr: Optional[ForeignProperty],
        left: Ref,
        right: DataType = None,
        *,
        # Used internally by swap method and only when fpr is None.
        _name: str = None,
        _chain: List[ForeignProperty] = None,
    ):
        if fpr is None:
            self.name = left.prop.place if _name is None else _name
            self.chain = [self] if _chain is None else (_chain + [self])
        else:
            self.name = fpr / left.prop
            self.chain = fpr.chain + [self]

        self.left = left
        self.right = right

    def __repr__(self):
        name = type(self).__name__
        model = self.chain[0].left.prop.model.name
        place = self / (self.right.prop if self.right else None)
        if self.right:
            return f'{name}({model}, {place}:{self.right.name})'
        else:
            return f'{name}({model}, {place})'

    def __truediv__(self, prop: Optional[Property]) -> str:
        if prop:
            return f'{self.name} -> {prop.place}'
        else:
            return f'{self.name} -> None'

    def join(self, fpr: ForeignProperty) -> ForeignProperty:
        """Join two ForeignProperty instances into one

        For example these two instances:

            ForeignProperty(City, country:Country -> None)
            ForeignProperty(Country, continent:Continent -> name:string)

        Would return:

            ForeignProperty(City, country:Country -> continent:Continent -> name:string)

        """
        tail = self
        head = fpr.chain[0]

        if tail.right is None:
            if tail.left.model.name != head.left.prop.model.name:
                raise IncompatibleForeignProperties(tail.left, right=head.left)
            tail = tail.push(head.left.prop)
        else:
            if tail.right != head.left:
                raise IncompatibleForeignProperties(tail.right, right=head.left)

        chain = list(tail.chain + head.chain[:-1])

        return ForeignProperty(
            None,
            fpr.left,
            fpr.right,
            _name=self / fpr.left.prop,
            _chain=chain,
        )

    def push(self, right: Optional[Property] = None) -> ForeignProperty:
        """Push another property to the chain

        Creates new ForeignProperty instance, by pushing another property to the
        chain. After this operation right will be pushed to left, and a new
        right will be placed instead of old one.
        """
        if isinstance(self.right, Ref):
            return ForeignProperty(self, self.right, right.dtype)
        elif self.right is None:
            return self.swap(right)
        else:
            raise RuntimeError(
                f"Can't push {right} to {self}, because right is a "
                f"{type(self.right)}, but it should be a Ref."
            )

    def swap(self, right: Optional[Property] = None) -> ForeignProperty:
        """Swap right with a new property

        Creates new ForeignProperty instance, preserving all properties, but
        change right to a new given right.
        """
        return ForeignProperty(
            None,
            self.left,
            right.dtype,
            _name=self.name,
            _chain=list(self.chain[:-1]),
        )

    def get_bind_expr(self) -> Expr:
        if self.right:
            expr = self.right.get_bind_expr()
        else:
            expr = None
        for fpr in reversed(self.chain[:-1]):
            if expr:
                expr = Expr('getattr', fpr.right.get_bind_expr(), expr)
            else:
                fpr.right.get_bind_expr()
        if expr:
            return Expr('getattr', self.chain[0].left.get_bind_expr(), expr)
        else:
            return self.chain[0].left.get_bind_expr()



from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.components import Context
from spinta.types.datatype import BackRef


@commands.prepare.register(Context, PostgreSQL, BackRef)
def prepare(context: Context, backend: PostgreSQL, dtype: BackRef):
    pass



import logging
from typing import List
from typing import Optional

from typer import Argument
from typer import Context as TyperContext
from typer import Exit
from typer import Option
from typer import echo

from spinta import commands
from spinta.cli.helpers.store import load_store
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Mode
from spinta.core.context import configure_context

log = logging.getLogger(__name__)


def run(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
    mode: Mode = Option('internal', help="Mode of backend operation"),
    host: str = Option('127.0.0.1', help="Run server on given host"),
    port: int = Option(8000, help="Run server on given port"),
    backend: Optional[str] = Option(None, '-b', '--backend', help=(
        "Backend connection string"
    )),
):
    """Run development server"""
    import os
    import uvicorn
    import spinta.api

    os.environ['AUTHLIB_INSECURE_TRANSPORT'] = '1'

    context = configure_context(ctx.obj, manifests, mode=mode, backend=backend)
    prepare_manifest(context, ensure_config_dir=True)
    app = spinta.api.init(context)

    echo("Spinta has started!")
    uvicorn.run(app, host=host, port=port)


def wait(
    ctx: TyperContext,
    seconds: Optional[int] = Argument(None),
):
    """Wait while all backends are up"""
    context = ctx.obj
    store = load_store(context)
    if not commands.wait(context, store, seconds=seconds, verbose=True):
        raise Exit(code=1)



from typing import Dict, Any

from spinta import commands
from spinta.components import Context
from spinta.backends.memory.components import Memory


@commands.load.register(Context, Memory, dict)
def load(context: Context, backend: Memory, config: Dict[str, Any]):
    backend.data = {}



from typing import Dict
from typing import Optional

from spinta.components import Property
from spinta.types.datatype import DataType


class Text(DataType):
    schema = {
        'langs': {'type': 'object'},
        'hidden': {'type': 'bool', 'default': True}
    }
    requires_source = False
    langs: Optional[Dict[str, Property]]



from typing import Any
from typing import Dict

from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.notimpl.components import BackendNotImplemented


@commands.load.register(Context, BackendNotImplemented, dict)
def load(
    context: Context,
    backend: BackendNotImplemented,
    config: Dict[str, Any],
) -> None:
    pass



from spinta import commands
from spinta.components import Context, DataSubItem
from spinta.backends.postgresql.components import PostgreSQL
from spinta.types.geometry.components import Geometry
from spinta.utils.data import take


@commands.before_write.register(Context, Geometry, PostgreSQL)
def before_write(
    context: Context,
    dtype: Geometry,
    backend: PostgreSQL,
    *,
    data: DataSubItem
):
    value = data.patch
    if (
        dtype.srid is not None and
        value and
        "SRID" not in value
    ):
        value = f"SRID={dtype.srid};{value}"
    return take({dtype.prop.place: value})



import calendar
import datetime
import sqlalchemy as sa

from spinta.backends.helpers import get_table_name
from spinta.core.ufuncs import Expr
from spinta.types.datatype import Integer, Number, Boolean, String, Date, DateTime, Time, Ref
from spinta import commands
from spinta.components import Context, Property
from spinta.components import Model
from spinta.exceptions import NotFoundError, NotImplementedFeature, InvalidRequestQuery
from spinta.exceptions import ItemDoesNotExist
from spinta.backends.postgresql.components import PostgreSQL
from spinta.types.geometry.components import Geometry
from spinta.ufuncs.summaryenv.components import SummaryEnv, BBox
from spinta.units.helpers import split_time_unit
from spinta.utils.nestedstruct import flat_dicts_to_nested

from dateutil.relativedelta import relativedelta


@commands.summary.register(Context, Model, PostgreSQL, Expr)
def summary(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    query: Expr
):
    env = SummaryEnv(context)
    env = env.init(model)
    env.resolve(query)
    kwargs = {}
    if env.bbox:
        kwargs['bbox'] = env.bbox
    result = commands.summary(context, env.prop.dtype, backend, **kwargs)
    return result


@commands.summary.register(Context, Integer, PostgreSQL)
def summary(
    context: Context,
    dtype: Integer,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection
    yield from _handle_numeric_summary(connection, dtype.prop)


@commands.summary.register(Context, Number, PostgreSQL)
def summary(
    context: Context,
    dtype: Number,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection
    yield from _handle_numeric_summary(connection, dtype.prop)


def _handle_numeric_summary(connection, model_prop: Property):
    try:
        prop = model_prop.name
        model = get_table_name(model_prop)
        min_max = connection.execute(f'SELECT MIN("{prop}") , MAX("{prop}") FROM "{model}"')
        min_value = 0
        max_value = 0
        for item in min_max:
            min_value = item[0]
            max_value = item[1]
        if min_value is None and max_value is None:
            return []
        else:
            if min_value == max_value:
                max_value += 100
            bin_size = (max_value - min_value) / 100
            half_bin_size = bin_size / 2

            # reason for 'max_value + min(0.01, bin_size*0.01)', is that WIDTH_BUCKET max is not inclusive, so you need to increase it by small amount
            result = connection.execute(f'''
                SELECT 
                    WIDTH_BUCKET("{prop}", {min_value}, {max_value + min(0.01, bin_size * 0.01)}, 100) AS bucket, 
                    COUNT(*) AS count,
                    (ARRAY_AGG(_id))[1] AS _id
                FROM "{model}" 
                GROUP BY 
                    bucket 
                ORDER BY 
                    bucket;
            ''')
            buckets = []
            for i in range(100):
                buckets.append(
                    {
                        "bin": i * bin_size + half_bin_size + min_value,
                        "count": 0,
                        "_type": model_prop.model.model_type()
                    }
                )
            for item in result:
                data = flat_dicts_to_nested(dict(item))
                if data["bucket"]:
                    bucket = buckets[data["bucket"] - 1]
                    bucket["count"] = data["count"]
                    if data['count'] == 1:
                        bucket["_id"] = data["_id"]
            yield from buckets
    except NotFoundError:
        raise ItemDoesNotExist(model_prop.model, id=model_prop.name)


@commands.summary.register(Context, Boolean, PostgreSQL)
def summary(
    context: Context,
    dtype: Boolean,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection
    try:
        prop = dtype.prop.name
        model = get_table_name(dtype.prop)

        result = connection.execute(f'''
                SELECT 
                    bin, 
                    COUNT(model.*) AS count, 
                    (ARRAY_AGG(model._id))[1] AS _id
                FROM UNNEST(ARRAY[TRUE, FALSE]) AS bin
                    LEFT OUTER JOIN 
                        "{model}" AS model 
                    ON 
                        bin = "{prop}"
                    GROUP BY 
                        bin
                    ORDER BY
                        bin;
        ''')
        for item in result:
            data = flat_dicts_to_nested(dict(item))
            if data["count"] != 1:
                del data["_id"]
            data['_type'] = dtype.prop.model.model_type()
            yield data
    except NotFoundError:
        raise ItemDoesNotExist(dtype.prop.model, id=dtype.prop.name)


@commands.summary.register(Context, String, PostgreSQL)
def summary(
    context: Context,
    dtype: String,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection

    if not dtype.prop.enum:
        raise NotImplementedFeature(dtype.prop,
                                    feature="Ability to generate summary for String type properties that do not have Enum")
    enum_list = []
    for key, value in dtype.prop.enum.items():
        enum_list.append(value.prepare)
    try:
        prop = dtype.prop.name
        model = get_table_name(dtype.prop)

        result = connection.execute(f'''
                SELECT 
                    bin, 
                    COUNT(model.*) AS count, 
                    (ARRAY_AGG(model._id))[1] AS _id
                FROM UNNEST(ARRAY{enum_list}) AS bin
                    LEFT OUTER JOIN 
                        "{model}" AS model 
                    ON 
                        bin = "{prop}"
                    GROUP BY 
                        bin
                    ORDER BY
                        bin;
                ''')
        for item in result:
            data = flat_dicts_to_nested(dict(item))
            if data["count"] != 1:
                del data["_id"]
            data['_type'] = dtype.prop.model.model_type()
            yield data
    except NotFoundError:
        raise ItemDoesNotExist(dtype.prop.model, id=dtype.prop.name)


@commands.summary.register(Context, Date, PostgreSQL)
def summary(
    context: Context,
    dtype: Date,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection
    yield from _handle_time_summary(connection, dtype.prop)


@commands.summary.register(Context, DateTime, PostgreSQL)
def summary(
    context: Context,
    dtype: DateTime,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection
    yield from _handle_time_summary(connection, dtype.prop)


@commands.summary.register(Context, Time, PostgreSQL)
def summary(
    context: Context,
    dtype: Time,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection
    yield from _handle_time_summary(connection, dtype.prop)


def _handle_time_units(model_prop: Property, value):
    unit = model_prop.unit

    if not unit:
        return _handle_time_units_not_given(model_prop, value)
    else:
        given = _handle_time_units_given(model_prop, value)
        if not given:
            return _handle_time_units_not_given(model_prop, value)
        if isinstance(model_prop.dtype, Time):
            if given.date() != value.date():
                return _handle_time_units_not_given(model_prop, value)
        return given


def _handle_time_units_not_given(model_prop: Property, value):
    if isinstance(model_prop.dtype, (Date, DateTime)):
        return value + datetime.timedelta(days=100)
    else:
        if (value + datetime.timedelta(minutes=100)).hour < value.hour:
            if (value + datetime.timedelta(seconds=100)).hour < value.hour:
                if (value + datetime.timedelta(milliseconds=100)).hour < value.hour:
                    return value + datetime.timedelta(microseconds=100)
                return value + datetime.timedelta(milliseconds=100)
            return value + datetime.timedelta(seconds=100)
        return value + datetime.timedelta(minutes=100)


def _handle_time_units_given(model_prop: Property, value):
    items = split_time_unit(model_prop.unit)
    if items is None:
        return None
    number = items[0]
    unit_type = items[1]
    time_units = ['H', 'T', 'S', 'L', 'U', 'N']
    date_units = ['Y', 'Q', 'M', 'W', 'D']
    if isinstance(model_prop.dtype, Time) and unit_type not in time_units:
        return None
    elif isinstance(model_prop.dtype, Date) and unit_type not in date_units:
        return None

    if unit_type == 'Y':
        return value + relativedelta(years=int(number) * 100)
    elif unit_type == 'Q':
        return value + relativedelta(months=int(number) * 3 * 100)
    elif unit_type == 'M':
        return value + relativedelta(months=int(number) * 100)
    elif unit_type == 'W':
        return value + relativedelta(weeks=int(number) * 100)
    elif unit_type == 'D':
        return value + relativedelta(days=int(number) * 100)
    elif unit_type == 'H':
        return value + relativedelta(hours=int(number) * 100)
    elif unit_type == 'T':
        return value + relativedelta(minutes=int(number) * 100)
    elif unit_type == 'S':
        return value + relativedelta(seconds=int(number) * 100)
    elif unit_type == 'L':
        return value + datetime.timedelta(milliseconds=int(number) * 100)
    elif unit_type == 'U':
        return value + datetime.timedelta(microseconds=int(number) * 100)


def _handle_time_summary(connection, model_prop: Property):
    try:
        prop = model_prop.name
        model = get_table_name(model_prop)
        min_value = None
        max_value = None
        if isinstance(model_prop.dtype, (Date, DateTime)):
            min_max = connection.execute(f'SELECT MIN("{prop}"::TIMESTAMP) , MAX("{prop}"::TIMESTAMP) FROM "{model}"')
            for item in min_max:
                min_value = item[0]
                max_value = item[1]
        else:
            min_max = connection.execute(f'SELECT MIN("{prop}") , MAX("{prop}") FROM "{model}"')
            for item in min_max:
                # 1970-01-01 is when postgresql start counting EPOCH
                if item[0] and item[1]:
                    min_value = datetime.datetime.combine(datetime.datetime(1970, 1, 1), item[0])
                    max_value = datetime.datetime.combine(datetime.datetime(1970, 1, 1), item[1])
        if min_value is None and max_value is None:
            return []
        else:
            if min_value == max_value:
                max_value = _handle_time_units(model_prop, min_value)

            bin_size = (max_value - min_value) / 100
            half_bin_size = bin_size / 2
            seq_start = calendar.timegm(min_value.timetuple())
            seq_end = calendar.timegm(max_value.timetuple())
            seq_step = bin_size.total_seconds()

            # reason for 'max_value + min(0.01, bin_size*0.01)', is that WIDTH_BUCKET max is not inclusive, so you need to increase it by small amount
            result = connection.execute(f'''
                SELECT 
                        WIDTH_BUCKET(EXTRACT(EPOCH FROM "{prop}"), {seq_start}, {seq_end + min(0.01, seq_step * 0.01)}, 100) AS bucket, 
                        COUNT(*) AS count,
                        (ARRAY_AGG(_id))[1] AS _id
                    FROM "{model}" 
                    GROUP BY 
                        bucket 
                    ORDER BY 
                        bucket;
                        ''')
            buckets = []
            for i in range(100):
                bucket_bin = i * bin_size + half_bin_size + min_value
                if isinstance(model_prop.dtype, Time):
                    bucket_bin = bucket_bin.time()
                buckets.append(
                    {
                        "bin": str(bucket_bin),
                        "count": 0,
                        "_type": model_prop.model.model_type()
                    }
                )
            for item in result:
                data = flat_dicts_to_nested(dict(item))
                if data["bucket"]:
                    bucket = buckets[data["bucket"] - 1]
                    bucket["count"] = data["count"]
                    if data['count'] == 1:
                        bucket["_id"] = data["_id"]
            yield from buckets
    except NotFoundError:
        raise ItemDoesNotExist(model_prop.model, id=model_prop.name)


@commands.summary.register(Context, Ref, PostgreSQL)
def summary(
    context: Context,
    dtype: Ref,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection

    try:
        prop = dtype.prop.name
        key = "_id"
        if dtype.prop.level and dtype.prop.level < 4:
            if len(dtype.refprops) > 1:
                raise NotImplementedFeature(dtype.prop, feature="Ability to get summary for Ref type Property, when level is 3 and below and there are multiple refprops")
            key = dtype.refprops[0].name
        model = get_table_name(dtype.prop)
        uri = dtype.model.uri
        prefixes = dtype.model.external.dataset.prefixes
        label = None
        if uri and ":" in uri:
            if uri.startswith(('http://', 'https://')):
                label = uri
            else:
                split = uri.split(":")
                if split[0] in prefixes.keys():
                    label = f'{prefixes[split[0]].uri}{split[1]}'

        result = connection.execute(f'''
                SELECT 
                    "{prop}.{key}" as bin, 
                    COUNT(model.*) AS count, 
                    (ARRAY_AGG(model._id))[1] AS _id,
                    MIN(model._created) as created_at
                FROM
                    "{model}" AS model 
                    GROUP BY 
                        "{prop}.{key}"
                    ORDER BY
                        count DESC,
                        created_at ASC
                    LIMIT 10;
                ''')
        for item in result:
            data = flat_dicts_to_nested(dict(item))

            del data["created_at"]
            if key == "_id":
                data["bin"] = extract_uuid(str(data["bin"]))
            if label:
                data["label"] = label
            if data["count"] != 1:
                del data["_id"]
            data['_type'] = dtype.prop.model.model_type()
            yield data
    except NotFoundError:
        raise ItemDoesNotExist(dtype.prop.model, id=dtype.prop.name)


def extract_uuid(uuid: str) -> str:
    return uuid[:8]


@commands.summary.register(Context, Geometry, PostgreSQL)
def summary(
    context: Context,
    dtype: Geometry,
    backend: PostgreSQL,
    **kwargs
):
    connection = context.get('transaction').connection
    try:
        prop = dtype.prop.name
        model = get_table_name(dtype.prop)
        bounding_box = ""
        params = {}
        if "bbox" in kwargs:
            bbox = kwargs["bbox"]

            if isinstance(bbox, BBox):
                if dtype.srid:
                    bounding_box = f'WHERE ST_Intersects("{prop}", ST_MakeEnvelope(:x_min, :y_min, :x_max, :y_max, :srid))'
                else:
                    bounding_box = f'WHERE ST_Intersects("{prop}", ST_MakeEnvelope(:x_min, :y_min, :x_max, :y_max))'
            else:
                raise InvalidRequestQuery(query="bbox", format="bbox(min_lon, min_lat, max_lon, max_lat)")

            params = {
                'x_min': bbox.x_min,
                'y_min': bbox.y_min,
                'x_max': bbox.x_max,
                'y_max': bbox.y_max,
                'srid': dtype.srid
            }

        count_exec = connection.execute(sa.text(f'SELECT COUNT(DISTINCT "{prop}") FROM "{model}" {bounding_box}'), params)
        count = 0
        for item in count_exec:
            count = item[0]
        params['count'] = count
        result = connection.execute(sa.text(f'''
                WITH clusters AS (
                    SELECT 
                    ST_ClusterKMeans(
                        model."{prop}", 
                        :count
                    ) OVER() AS cluster_id,
                    model."{prop}" AS geom,
                    model._id AS _id,
                    model._created as created_at
                    FROM "{model}" AS model
                    {bounding_box}
                )
                SELECT 
                    ST_NumGeometries(ST_Collect(geom)) as cluster,
                    ST_AsText(ST_Centroid(ST_Collect(geom))) AS centroid,
                    (ARRAY_AGG(clusters._id))[1] AS _id
                FROM clusters
                GROUP BY cluster_id
                ORDER BY MIN(clusters.created_at);
                '''), params)
        for item in result:
            data = flat_dicts_to_nested(dict(item))
            if data["cluster"]:
                if data["cluster"] != 1:
                    del data["_id"]
                data['_type'] = dtype.prop.model.model_type()
                yield data
    except NotFoundError:
        raise ItemDoesNotExist(dtype.prop.model, id=dtype.prop.name)



from spinta.components import Model
from spinta.core.ufuncs import Env
from spinta.exceptions import UnknownRequestQuery


class BBox:
    x_min: float
    y_min: float
    x_max: float
    y_max: float

    def __init__(self, x_min: float, y_min: float, x_max: float, y_max: float,):
        self.x_min = x_min
        self.y_min = y_min
        self.x_max = x_max
        self.y_max = y_max


class SummaryEnv(Env):

    def init(self, model: Model):
        return self(
            model=model,
            prop=None,
            bbox=None
        )

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownRequestQuery(request="summary", query=expr.name)



import logging

from typing import Optional

from spinta import exceptions
from spinta.typing import ObjectData

log = logging.getLogger(__name__)


def report_error(
    exc: Exception,
    data: Optional[ObjectData] = None,
    *,
    stop_on_error: bool = True,
) -> Exception:
    if data and '_id' in data and isinstance(exc, exceptions.BaseError):
        exc.context['id'] = data['_id']
    if stop_on_error:
        raise exc
    else:
        log.error(str(exc))
    return exc



from typing import Any
from typing import Iterator
from typing import Optional

from starlette.requests import Request
from starlette.responses import Response

from spinta.components import Context, Action, UrlParams, Node
from spinta import commands


def render(
    context: Context,
    request: Request,
    node: Node,
    params: UrlParams,
    data: Iterator[Any],
    *,
    action: Optional[Action] = None,
    status_code: Optional[int] = 200,
    headers: Optional[dict] = None,
) -> Response:
    return commands.render(
        context, request, node, params.fmt,
        action=action,
        params=params,
        data=data,
        status_code=status_code,
        headers=headers,
    )



from types import AsyncGeneratorType
from typing import Optional

import sqlalchemy as sa

from spinta import commands
from spinta.utils.json import fix_data_for_json
from spinta.components import Context, Action, Model, Property
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.sqlalchemy import utcnow


@commands.create_changelog_entry.register(Context, (Model, Property), PostgreSQL)
async def create_changelog_entry(
    context: Context,
    node: (Model, Property),
    backend: PostgreSQL,
    *,
    dstream: AsyncGeneratorType,
) -> None:
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(node, TableType.CHANGELOG)
    async for data in dstream:

        if not data.patch:
            yield data
            continue

        qry = table.insert().values(
            _txn=transaction.id,
            datetime=utcnow(),
            action=Action.INSERT.value,
        )
        connection.execute(qry, [{
            '_rid': data.saved['_id'] if data.saved else data.patch['_id'],
            '_revision': data.patch['_revision'] if data.patch else data.saved['_revision'],
            '_txn': transaction.id,
            'datetime': utcnow(),
            'action': data.action.value,
            'data': fix_data_for_json({
                k: v for k, v in data.patch.items() if not k.startswith('_')
            }),
        }])
        yield data


@commands.changes.register(Context, Model, PostgreSQL)
def changes(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    id_: str = None,
    limit: int = 100,
    offset: Optional[int] = -100,
):
    connection = context.get('transaction').connection
    table = backend.get_table(model, TableType.CHANGELOG)

    qry = sa.select([table]).order_by(table.c['_id'].asc())
    qry = _changes_id(table, qry, id_)
    qry = _changes_offset(table, qry, offset)
    qry = _changes_limit(qry, limit)

    result = connection.execute(qry)
    for row in result:
        yield {
            '_cid': row[table.c['_id']],
            '_created': row[table.c['datetime']],
            '_op': row[table.c['action']],
            '_id': row[table.c['_rid']],
            '_txn': row[table.c['_txn']],
            '_revision': row[table.c['_revision']],
            **dict(row[table.c['data']]),
        }


def _changes_id(table, qry, id_):
    if id_:
        return qry.where(table.c['_rid'] == id_)
    else:
        return qry


def _changes_offset(table, qry, offset):
    if offset is not None:
        if offset >= 0:
            return qry.where(table.c['_id'] >= offset)
        else:
            offset = (
                qry.with_only_columns([
                    sa.func.max(table.c['_id']) - abs(offset),
                ]).
                order_by(None).
                scalar_subquery()
            )
            return qry.where(table.c['_id'] > offset)
    else:
        return qry


def _changes_limit(qry, limit):
    if limit:
        return qry.limit(limit)
    else:
        return qry



from typing import Dict, Any

from spinta import commands
from spinta.backends.nobackend.components import NoBackend
from spinta.components import Context


@commands.load.register(Context, NoBackend, dict)
def load(
    context: Context,
    backend: NoBackend,
    config: Dict[str, Any],
) -> None:
    pass



import datetime

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Date
from spinta.backends.mongo.components import Mongo


@commands.prepare_for_write.register(Context, Date, Mongo, datetime.date)
def prepare_for_write(
    context: Context,
    dtype: Date,
    backend: Mongo,
    value: datetime.date,
) -> datetime.datetime:
    # prepares date values for Mongo store, they must be converted to datetime
    return datetime.datetime.combine(value, datetime.datetime.min.time())


@commands.cast_backend_to_python.register(Context, Date, Mongo, datetime.datetime)
def cast_backend_to_python(
    context: Context,
    dtype: Date,
    backend: Mongo,
    value: datetime.datetime,
) -> datetime.date:
    return value.date()



from spinta import commands
from spinta.components import Context
from spinta.components import Model
from spinta.backends.nobackend.components import NoBackend


@commands.wipe.register(Context, Model, NoBackend)
def wipe(context: Context, model: Model, backend: NoBackend):
    pass



from operator import itemgetter
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Tuple

import sqlalchemy as sa
from geoalchemy2.types import Geometry
from sqlalchemy.dialects import mysql
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects import oracle
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.types import TypeEngine

from spinta import spyna
from spinta.components import Context
from spinta.core.ufuncs import asttoexpr
from spinta.datasets.backends.sql.ufuncs.components import SqlResource, Engine
from spinta.exceptions import UnexpectedFormulaResult
from spinta.utils.imports import full_class_name
from spinta.utils.naming import Deduplicator
from spinta.utils.naming import to_dataset_name
from spinta.utils.naming import to_model_name
from spinta.utils.naming import to_property_name


def read_schema(context: Context, path: str, prepare: str = None):
    engine = sa.create_engine(path)
    schema = None
    if prepare:
        env = SqlResource(context).init(path)
        parsed = spyna.parse(prepare)
        converted = asttoexpr(parsed)
        engine = env.resolve(converted)
        engine = env.execute(engine)
        if not isinstance(engine, Engine):
            raise UnexpectedFormulaResult(
                formula=spyna.unparse(converted),
                expected=full_class_name(Engine),
                received=full_class_name(engine),
            )

        schema = engine.schema
        engine = engine.create()

    url = sa.engine.make_url(path)
    dataset = to_dataset_name(url.database) if url.database else 'dataset1'
    insp = sa.inspect(engine)

    table_mapper = [
        {
            "dataset": dataset,
            "resource": 'resource1',
            "mapping": _create_mapping(insp, insp.get_table_names(schema=schema), schema, dataset)
        }
    ]

    get_view_names = getattr(insp, "get_view_names", None)
    get_materialized_view_names = getattr(insp, "get_materialized_view_names", None)
    views = []
    if callable(get_view_names):
        views += insp.get_view_names(schema=schema)
    if callable(get_materialized_view_names):
        views += insp.get_materialized_view_names(schema=schema)

    if views:
        dataset = f'{dataset}/views'
        table_mapper.append(
            {
                "dataset": dataset,
                "resource": "resource1",
                "mapping": _create_mapping(insp, views, schema, dataset)
            }
        )

    for mapping_data in table_mapper:
        yield None, {
            'type': 'dataset',
            'name': mapping_data["dataset"],
            'resources': {
                mapping_data["resource"]: {
                    'type': 'sql',
                    'external': str(url.set(password='')),
                    'prepare': prepare
                },
            },
        }

        for table in sorted(mapping_data["mapping"]):
            yield None, {
                'type': 'model',
                'name': mapping_data["mapping"][table].model,
                'external': {
                    'dataset': mapping_data["dataset"],
                    'resource': mapping_data["resource"],
                    'name': table,
                    'pk': _get_primary_key(insp, table, schema, mapping_data["mapping"]),
                },
                'description': _get_table_comment(insp, schema, table),
                'properties': dict(_read_props(insp, table, schema, mapping_data["mapping"])),
            }


class _TableMapping(NamedTuple):
    model: str  # full model name
    props: Dict[
        str,    # column
        str,    # property
    ]


_Mapping = Dict[
    str,  # table
    _TableMapping,
]


def _create_mapping(insp: Inspector, tables: list, schema: str, dataset: str) -> _Mapping:
    dedup_model = Deduplicator('{}')
    mapping: _Mapping = {}
    for table in sorted(tables):
        model = to_model_name(table)
        model = dedup_model(model)
        dedup_prop = Deduplicator('_{}')
        props = {}
        for col in insp.get_columns(table, schema=schema):
            prop = to_property_name(col['name'])
            prop = dedup_prop(prop)
            props[col['name']] = prop
        mapping[table] = _TableMapping(
            dataset + '/' + model,
            props,
        )
    return mapping


def _get_primary_key(
    insp: Inspector,
    table: str,
    schema: str,
    mapping: _TableMapping,
) -> List[str]:
    pk = insp.get_pk_constraint(table, schema=schema)
    return [
        mapping[table].props[col]
        for col in pk['constrained_columns']
    ]


def _get_table_comment(insp: Inspector, schema: str, table: str) -> str:
    try:
        return insp.get_table_comment(table, schema=schema).get('text')
    except NotImplementedError:
        return ''


def _read_props(
    insp: Inspector,
    table: str,
    schema: str,
    mapping: _Mapping,
) -> Iterator[Tuple[
    str,
    Dict[str, Any],
]]:
    fkeys, cfkeys = _get_fkeys(insp, table, schema, mapping)

    cols = insp.get_columns(table, schema=schema)
    cols = sorted(cols, key=itemgetter('name'))
    for col in cols:
        name = col['name']
        prop = mapping[table].props[name]
        extra = {}

        if name in cfkeys:
            ref = cfkeys[name]
            yield ref.name, {
                'type': 'ref',
                'model': ref.model,
                'refprops': ref.props,
            }

        if name in fkeys:
            ref = fkeys[name]
            dtype = 'ref'
            extra = {
                'model': ref.model,
                'refprops': ref.props,
            }

        else:
            dtype = _get_type(col['type'])

        yield prop, {
            'type': dtype,
            'external': {
                'name': name,
            },
            'description': col.get('comment') or '',
            **extra,
        }


TYPES = [
    (sa.Boolean, 'boolean'),
    (sa.Date, 'date'),
    (sa.DateTime, 'datetime'),
    (sa.Float, 'number'),
    (sa.Integer, 'integer'),
    (sa.Numeric, 'number'),
    (sa.Text, 'string'),
    (sa.Time, 'time'),
    (sa.LargeBinary, 'binary'),
    (sa.String, 'string'),
    (sa.VARCHAR, 'string'),
    (sa.CHAR, 'string'),
    (mysql.BIT, 'string'),
    (mysql.VARBINARY, 'string'),
    (mysql.VARCHAR, 'string'),
    (postgresql.ARRAY, 'array'),
    (postgresql.JSON, 'object'),
    (postgresql.JSONB, 'object'),
    (postgresql.UUID, 'string'),
    (postgresql.INTERVAL, 'integer'),  # total number of seconds
    (Geometry, 'geometry'),
    (oracle.ROWID, 'string'),
    (oracle.RAW, 'binary'),
]


def _get_type(sql_type: TypeEngine) -> str:
    for cls, name in TYPES:
        if isinstance(sql_type, cls):
            return name
    raise TypeError(f"Unknown type {sql_type!r}.")


class _Ref(NamedTuple):
    name: str    # property name
    model: str   # referenced model
    props: List[str]  # referenced props


def _get_fkeys(
    insp: Inspector,
    table: str,
    schema: str,
    mapping: _Mapping,
) -> Tuple[
    Dict[       # foreign keys
        str,    # column (source)
        _Ref,
    ],
    Dict[       # composite foreign keys
        str,    # column (source)
        _Ref,
    ],
]:
    fkeys = {}
    cfkeys = {}
    for fk in insp.get_foreign_keys(table, schema=schema):
        composite = len(fk['constrained_columns']) > 1

        col = fk['constrained_columns'][0]

        rtable = fk['referred_table']

        if composite:
            name = '_'.join([
                mapping[table].props[c]
                for c in fk['constrained_columns']
            ])
        else:
            name = mapping[table].props[col]

        ref = _Ref(
            name=name,
            model=mapping[rtable].model,
            props=[
                mapping[rtable].props[rcol]
                for rcol in fk['referred_columns']
            ],
        )

        if composite:
            cfkeys[col] = ref
        else:
            fkeys[col] = ref

    return fkeys, cfkeys



import asyncio
import pathlib
from typing import List
from typing import Optional

from typer import Context as TyperContext
from typer import Exit
from typer import Option
from typer import echo

from spinta import commands
from spinta import exceptions
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import process_stream
from spinta.cli.helpers.store import prepare_manifest
from spinta.commands.write import write
from spinta.components import Context
from spinta.components import Model
from spinta.datasets.components import Dataset
from spinta.manifests.components import Manifest


def _get_dataset_models(manifest: Manifest, dataset: Dataset):
    for model in manifest.models.values():
        if model.external and model.external.dataset and model.external.dataset.name == dataset.name:
            yield model


def _pull_models(context: Context, models: List[Model]):
    for model in models:
        external = model.external
        backend = external.resource.backend
        rows = commands.getall(context, external, backend)
        yield from rows


def pull(
    ctx: TyperContext,
    dataset: str,
    model: Optional[List[str]] = Option(None, '-m', '--model', help=(
        "Pull only specified models."
    )),
    push: bool = Option(False, help="Write pulled data to database"),
    export: str = Option(None, '-e', '--export', help=(
        "Export pulled data to a file or stdout. For stdout use "
        "'stdout:<fmt>', where <fmt> can be 'csv' or other supported format."
    )),
):
    """Pull data from an external data source."""
    context = ctx.obj
    store = prepare_manifest(context)
    manifest = store.manifest
    if dataset in manifest.objects['dataset']:
        dataset = manifest.objects['dataset'][dataset]
    else:
        echo(str(exceptions.NodeNotFound(manifest, type='dataset', name=dataset)))
        raise Exit(code=1)

    if model:
        models = []
        for model in model:
            if model not in manifest.models:
                echo(str(exceptions.NodeNotFound(manifest, type='model', name=model)))
                raise Exit(code=1)
            models.append(manifest.models[model])
    else:
        models = _get_dataset_models(manifest, dataset)

    try:
        with context:
            require_auth(context)
            backend = store.backends['default']
            context.attach('transaction', backend.transaction, write=push)

            path = None
            exporter = None

            stream = _pull_models(context, models)
            if push:
                root = manifest.objects['ns']['']
                stream = write(context, root, stream, changed=True)

            if export is None and push is False:
                export = 'stdout'

            if export:
                if export == 'stdout':
                    fmt = 'ascii'
                elif export.startswith('stdout:'):
                    fmt = export.split(':', 1)[1]
                else:
                    path = pathlib.Path(export)
                    fmt = export.suffix.strip('.')

                config = context.get('config')

                if fmt not in config.exporters:
                    echo(f"unknown export file type {fmt!r}")
                    raise Exit(code=1)

                exporter = config.exporters[fmt]

            asyncio.get_event_loop().run_until_complete(
                process_stream(dataset.name, stream, exporter, path)
            )

    except exceptions.BaseError as e:
        echo(str(e))
        raise Exit(code=1)



from typing import overload
from typing import Iterator

from spinta.typing import ObjectData
from spinta import commands
from spinta.core.ufuncs import Expr
from spinta.components import Context, UrlParams
from spinta.components import Model
from spinta.exceptions import NotFoundError
from spinta.exceptions import ItemDoesNotExist
from spinta.backends.postgresql.components import PostgreSQL
from spinta.ufuncs.basequerybuilder.components import get_page_values, QueryParams
from spinta.ufuncs.helpers import merge_formulas
from spinta.utils.nestedstruct import flat_dicts_to_nested
from spinta.backends.postgresql.commands.query import PgQueryBuilder


@overload
@commands.getone.register(Context, Model, PostgreSQL)
def getone(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    id_: str,
) -> ObjectData:
    connection = context.get('transaction').connection
    table = backend.get_table(model)
    try:
        result = backend.get(connection, table, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(model, id=id_)
    data = flat_dicts_to_nested(dict(result))
    data['_type'] = model.model_type()
    return commands.cast_backend_to_python(context, model, backend, data)


@commands.getall.register(Context, Model, PostgreSQL)
def getall(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    query: Expr = None,
    default_expand: bool = True,
    params: QueryParams = None,
    **kwargs
) -> Iterator[ObjectData]:
    assert isinstance(query, (Expr, type(None))), query
    connection = context.get('transaction').connection

    if default_expand:
        query = merge_formulas(
            query, Expr('expand')
        )

    builder = PgQueryBuilder(context)
    builder.update(model=model)
    table = backend.get_table(model)
    env = builder.init(backend, table, params)
    expr = env.resolve(query)
    where = env.execute(expr)
    qry = env.build(where)

    conn = connection.execution_options(stream_results=True)
    result = conn.execute(qry)

    for row in result:
        converted = flat_dicts_to_nested(dict(row))
        res = {
            '_type': model.model_type(),
            **converted
        }
        if model.page.is_enabled:
            res['_page'] = get_page_values(env, row)
        res = commands.cast_backend_to_python(context, model, backend, res)
        yield res
